/* minimal program to open & clear a window */
#include <cmath>
#include <iostream>
#include <GL/glew.h>
#include <GL/freeglut.h>
#include <GL/freeglut_ext.h>
#include "moglclass.h"
#include "shaders.h"
#include "vertexBufferObject.h"
#include "geometryPositionColor.h"
#include "simpleCar.h"
#include "simpleRobot.h"
#include "simpleSolar.h"
#include "simpleMobile.h"
#include "simpleGeo.h"
#include "simpleTree.h"
#include "simpleCube3d.h"
#include "camera.h"

using namespace moglclass;

#define NUMSHADERS 2
Shader shShaders[NUMSHADERS];
Program spMain;

int startTime = 0; // animation
SimpleCar *car1;
SimpleRobot *robot1;
SimpleSolar *solar1;
SimpleMobile *mobile1;
SimpleGeo *geo1;
SimpleTree *tree1;
SimpleCube3d *cube1;

bool g_mousemove = false;
bool g_wiremode = false;
bool g_animationmode = false;
int g_transformationmode = 0;

float g_fovy = 45.0f;
float g_aspect = 1.0f;
float g_zNear = 1.0f;
float g_zFar = 100.f;

Camera camera1(FLY);
glm::vec3 g_eye(0, 0, 10);
glm::vec3 g_at(0, 0, 0);
glm::vec3 g_up(0, 1, 0);

glm::mat4 World(1.0f);
glm::mat4 View(1.0f);
glm::mat4 Projection(1.0f);

float g_theta = 0.0f;
glm::mat4 g_spinTransform(1.0f);


////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////
void mprint(glm::mat4& Mat)
{
	printf("\n %f %f %f %f\n %f %f %f %f\n %f %f %f %f\n %f %f %f %f\n\n",
		Mat[0][0], Mat[1][0], Mat[2][0], Mat[3][0],
		Mat[0][1], Mat[1][1], Mat[2][1], Mat[3][1],
		Mat[0][2], Mat[1][2], Mat[2][2], Mat[3][2],
		Mat[0][3], Mat[1][3], Mat[2][3], Mat[3][3]);
}
// set geometry
void setData()
{
	car1 = new SimpleCar();
	solar1 = new SimpleSolar();
	mobile1 = new SimpleMobile();
	geo1 = new SimpleGeo();
	cube1 = new SimpleCube3d();
	robot1 = new SimpleRobot();
	tree1 = new SimpleTree();
}

void init( void )
{
    // Load shaders and use the resulting shader program
	spMain.createProgram();
	shShaders[0].loadShader("simple4.vs", GL_VERTEX_SHADER);
	shShaders[1].loadShader("simple4.fs", GL_FRAGMENT_SHADER);
	spMain.attachShaderToProgram(&shShaders[0]);
	spMain.attachShaderToProgram(&shShaders[1]);
	spMain.linkProgram();
	spMain.useProgram();

	// MVP matrix
	Projection = glm::perspective(g_fovy, g_aspect, g_zNear, g_zFar);
	//View = glm::lookAt(g_eye, g_at, g_up);
	View = camera1.lookAt(g_eye, g_at, g_up);

	setData();
	
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LESS);
	glClearColor( 1.0, 1.0, 1.0, 1.0 ); // white background
}

void display( void )
{
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT ); // clear the window

	// set drawmode
	if (g_wiremode) 
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);  // draw wireframe 
	else
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);	// draw filled 

	glm::mat4 carMatrix = World * glm::translate(glm::mat4(1.0f), glm::vec3(-6.0f, 0.0f, 0.0f));
	car1->draw(&spMain, Projection, View, carMatrix);
	glm::mat4 solarMatrix = World * glm::translate(glm::mat4(1.0f), glm::vec3(6.0f, 0.0f, 0.0f)) * glm::scale(glm::mat4(1.0f), glm::vec3(0.5f, 0.5f, 0.5f));
	solar1->draw(&spMain, Projection, View, solarMatrix);
	glm::mat4 mobileMatrix = World * glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, 0.0f, 6.0f)) * glm::scale(glm::mat4(1.0f), glm::vec3(0.3f, 0.3f, 0.3f));
	mobile1->draw(&spMain, Projection, View, mobileMatrix);
	glm::mat4 geoMatrix = World * glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, 0.0f, -6.0f)) * glm::scale(glm::mat4(1.0f), glm::vec3(0.5f, 0.5f, 0.5f));
	geo1->draw(&spMain, Projection, View, geoMatrix);
	glm::mat4 cubeMatrix = World * glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, -6.0f, 0.0f));
	cube1->draw(&spMain, Projection, View, cubeMatrix);
	//glm::mat4 robotMatrix = World * glm::translate(glm::mat4(1.0f), glm::vec3(-5.0f, 5.0f, 5.0f));
	//robot1->draw(&spMain, Projection, View, robotMatrix);
	//glm::mat4 treeMatrix = World * glm::translate(glm::mat4(1.0f), glm::vec3(5.0f, 5.0f, 5.0f));
	//tree1->draw(&spMain, Projection, View, treeMatrix);

	// MVP
	spMain.useProgram();
	spMain.setUniform("gProjection", Projection);
	spMain.setUniform("gView", View);	

	glutSwapBuffers();
}

void update()
{
	static int currentTime, deltaTime, prevTime = 0; 
	static int elapsedTime = 0;
	currentTime = glutGet(GLUT_ELAPSED_TIME);
	deltaTime = currentTime - prevTime;

	car1->update((float)deltaTime);
	solar1->update((float)deltaTime * 0.05f);
	mobile1->update((float)deltaTime * 0.05f);
	geo1->update((float)deltaTime);
	cube1->update((float)elapsedTime);

	//robot1->update((float)deltaTime * 0.01f);

	elapsedTime = currentTime - startTime;
	prevTime = currentTime;

	glutPostRedisplay();
}

void keyboard( unsigned char key, int x, int y )
{
	switch ( key ) {
		case 27:
			exit( EXIT_SUCCESS );
			break;
		case 'r':
			View = camera1.lookAt(g_eye, g_at, g_up);
			break;
		case 'w':
			g_wiremode = !g_wiremode;
			break;
		case 'a':
			g_animationmode = !g_animationmode;
			if (!car1->isActive()) car1->activate();
			else car1->deactivate();
			if (!solar1->isActive()) solar1->activate();
			else solar1->deactivate();
			if (!mobile1->isActive()) mobile1->activate();
			else mobile1->deactivate();
			if (!geo1->isActive()) geo1->activate();
			else geo1->deactivate();
			if (!cube1->isActive()) cube1->activate();
			else cube1->deactivate();
			//if (!robot1->isActive()) robot1->activate();
			//else robot1->deactivate();
			//if (!tree1->isActive()) tree1->activate();
			//else tree1->deactivate();
			startTime = glutGet(GLUT_ELAPSED_TIME);
			break;
		case 'p': // perspective projection
			Projection = glm::perspective(g_fovy, g_aspect, g_zNear, g_zFar);
			break;
		case 'o': // orthogonal projection
			Projection = glm::ortho(-5.0f, 5.0f, -5.0f, 5.0f, g_zNear, g_zFar);
			break;
	}
	glutPostRedisplay();
}


void specialkey(int key, int x, int y)
{
	if (key == GLUT_KEY_F1) { // x-movement
		camera1.strafe(0.5); 
		glm::vec3 pos = camera1.getPosition();
		printf("camera1 =%f %f %f\n", pos[0], pos[1], pos[2]);
	}
	else if (key == GLUT_KEY_F2) {
		camera1.strafe(-0.5);
		glm::vec3 pos = camera1.getPosition();
		printf("camera1 =%f %f %f\n", pos[0], pos[1], pos[2]);
	}
	else if (key == GLUT_KEY_F3) { // y-movement
		camera1.fly(0.5);
		glm::vec3 pos = camera1.getPosition();
		printf("camera1 =%f %f %f\n", pos[0], pos[1], pos[2]);
	}
	else if (key == GLUT_KEY_F4) {
		camera1.fly(-0.5);
		glm::vec3 pos = camera1.getPosition();
		printf("camera1 =%f %f %f\n", pos[0], pos[1], pos[2]);
	}
	else if (key == GLUT_KEY_F5) { // z-movement
		camera1.walk(0.5);
		glm::vec3 pos = camera1.getPosition();
		printf("camera1 =%f %f %f\n", pos[0], pos[1], pos[2]);
	}
	else if (key == GLUT_KEY_F6) {
		camera1.walk(-0.5);
		glm::vec3 pos = camera1.getPosition();
		printf("camera1 =%f %f %f\n", pos[0], pos[1], pos[2]);
	}

	else if (key == GLUT_KEY_F7) { // yaw (by y-axis)
		camera1.yaw(2.5);
		glm::vec3 right = camera1.getRight();
		glm::vec3 look = camera1.getLook();
		printf("camera1 YAW Right Vector =%f %f %f\n", right[0], right[1], right[2]);
		printf("camera1 YAW Look Vector =%f %f %f\n", look[0], look[1], look[2]);
	}
	else if (key == GLUT_KEY_F8) {
		camera1.yaw(-2.5);
		glm::vec3 right = camera1.getRight();
		glm::vec3 look = camera1.getLook();
		printf("camera1 YAW Right Vector =%f %f %f\n", right[0], right[1], right[2]);
		printf("camera1 YAW Look Vector =%f %f %f\n", look[0], look[1], look[2]);
	}
	else if (key == GLUT_KEY_F9) { // pitch (by x-axis)
		camera1.pitch(2.5);
		glm::vec3 up = camera1.getUp();
		glm::vec3 look = camera1.getLook();
		printf("camera1 PITCH Up Vector =%f %f %f\n", up[0], up[1], up[2]);
		printf("camera1 PITCH Look Vector =%f %f %f\n", look[0], look[1], look[2]);
	}
	else if (key == GLUT_KEY_F10) {
		camera1.pitch(-2.5);
		glm::vec3 up = camera1.getUp();
		glm::vec3 look = camera1.getLook();
		printf("camera1 PITCH Up Vector =%f %f %f\n", up[0], up[1], up[2]);
		printf("camera1 PITCH Look Vector =%f %f %f\n", look[0], look[1], look[2]);
	}
	else if (key == GLUT_KEY_HOME) { // roll (by z-axis)
		camera1.roll(2.5);
		glm::vec3 right = camera1.getRight();
		glm::vec3 up = camera1.getUp();
		printf("camera1 ROLL Right Vector =%f %f %f\n", right[0], right[1], right[2]);
		printf("camera1 ROLL Up Vector =%f %f %f\n", up[0], up[1], up[2]);
	}
	else if (key == GLUT_KEY_END) {
		camera1.roll(-2.5);
		glm::vec3 right = camera1.getRight();
		glm::vec3 up = camera1.getUp();
		printf("camera1 ROLL Right Vector =%f %f %f\n", right[0], right[1], right[2]);
		printf("camera1 ROLL Up Vector =%f %f %f\n", up[0], up[1], up[2]);
	}

	else if (key == GLUT_KEY_LEFT) // same as town
		camera1.yaw(2.5);
	else if (key == GLUT_KEY_RIGHT)
		camera1.yaw(-2.5);
	else if (key == GLUT_KEY_UP)
		camera1.walk(0.5);
	else if (key == GLUT_KEY_DOWN)
		camera1.walk(-0.5);

	View = camera1.View();
	glutPostRedisplay();
}


void mouse(int button, int state, int mx, int my)
{
	if (button==GLUT_RIGHT_BUTTON && state==GLUT_DOWN)
		exit(0);
	if (button==GLUT_LEFT_BUTTON && state==GLUT_DOWN)
		g_mousemove = true;
	else if (button==GLUT_LEFT_BUTTON && state==GLUT_UP)
		g_mousemove = false;
}

void motion(int mx, int my)
{
	int w = glutGet(GLUT_WINDOW_WIDTH);
	int h = glutGet(GLUT_WINDOW_HEIGHT);

	// 0~600(x+ right) x 0~600(y+ down) => -5~5(x+ right) x -5~5(y+ up)
	float x = (float)  10.0f * (mx - w*0.5f) / w;
	float y = (float) 10.0f * (h*0.5f - my) / h;

	if (g_mousemove)
	{
		//printf("mx=%d my=%d x=%f y=%f\n", mx, my, x, y);
		//geoList[g_geometrymode]->setPosition(glm::vec3(x, y, 0));
	}

	glutPostRedisplay();
}

void reshape(int w, int h)
{
	g_aspect = (float) (w/h);

    glViewport(0, 0, (GLsizei) w, (GLsizei) h);
    glutPostRedisplay();
}

void cleanup()
{		 
	spMain.deleteProgram();
}

int main( int argc, char **argv )
{ 
	glutInit(&argc, argv);
	glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGBA  | GLUT_DEPTH);
	glutInitWindowSize( 600, 600 );
	glutCreateWindow( "Geometry Primitives Orientation & Camera" );

	glewInit();
	init();
	glutDisplayFunc( display );
	glutKeyboardFunc( keyboard );
	glutSpecialFunc( specialkey );
	glutMouseFunc( mouse ); 
	glutMotionFunc( motion ); 
	glutIdleFunc( update );
	glutReshapeFunc( reshape );
	glutMainLoop();

	cleanup();
	return 0;
}